"""
Helper class to handle objects that contain a pointer to objects of varied types, usually all with the same base type.
"""
import copy
from typing import Dict, Union, Type

import construct
from construct import Construct, Struct, Hex, Int64ul, Computed, Switch, Adapter

from mercury_engine_data_structures import hashed_names
from mercury_engine_data_structures.construct_extensions.misc import ErrorWithMessage


class PointerAdapter(Adapter):
    types: Dict[int, Union[Construct, Type[Construct]]]

    def __init__(self, subcon, types):
        super().__init__(subcon)
        self.types = types

    @property
    def _allow_null(self):
        return hashed_names.all_name_to_property_id()["void"] in self.types

    @property
    def _single_type(self):
        return len(self.types) == (2 if self._allow_null else 1)

    def _decode(self, obj: construct.Container, context, path):
        if obj.ptr is None:
            return None

        if self._single_type:
            return obj.ptr

        ret = construct.Container()
        ret["@type"] = hashed_names.all_property_id_to_name()[obj.type]
        for key, value in obj.ptr.items():
            ret[key] = value
        return ret

    def _encode(self, obj: construct.Container, context, path):
        if obj is None:
            type_id = hashed_names.all_name_to_property_id()["void"]

        elif self._single_type:
            type_id = list(self.types.keys())[1]

        else:
            obj = copy.copy(obj)
            type_name: str = obj.pop("@type")
            type_id = hashed_names.all_name_to_property_id()[type_name]

        return construct.Container(
            type=type_id,
            ptr=obj,
        )


class PointerSet:
    types: Dict[int, Union[Construct, Type[Construct]]]

    def __init__(self, category: str, *, allow_null: bool = True):
        self.category = category
        self.types = {}
        if allow_null:
            self.add_option("void", construct.Pass)

    @classmethod
    def construct_pointer_for(cls, name: str, conn: Union[Construct, Type[Construct]]) -> Construct:
        ret = cls(name, allow_null=True)
        ret.add_option(name, conn)
        return ret.create_construct()

    def add_option(self, name: str, value: Union[Construct, Type[Construct]]) -> None:
        prop_id = hashed_names.all_name_to_property_id()[name]
        if prop_id in self.types:
            raise ValueError(f"Attempting to add {name} to {self.category}, but already present.")
        self.types[prop_id] = name / value

    def create_construct(self) -> Construct:
        return PointerAdapter(Struct(
            type=Hex(Int64ul),
            ptr=Switch(
                construct.this.type,
                self.types,
                ErrorWithMessage(
                    lambda ctx: f"Property {ctx.type} ({hashed_names.all_property_id_to_name().get(ctx.type)}) "
                                "without assigned type"),
            )
        ), self.types)



from construct import (
    Struct, Construct, Const, Int32ul, PrefixedArray, CString, Byte, Array, Float32l, PaddedString,
    Int64ul
)

from mercury_engine_data_structures.formats import BaseResource
from mercury_engine_data_structures.game_check import Game

VectorArray = Array(3, Float32l)

BMSSD = Struct(
    magic=Const(b"MSSD"),
    unk1=Int32ul,
    part_info=PrefixedArray(
        Int32ul,
        Struct(
            model_name=CString("utf-8"),
            byte0=Byte,
            byte1=Byte,
            byte2=Byte,
            int3=Int32ul,
            byte4=Byte,
            farr4=VectorArray,
            farr5=VectorArray,
            farr6=VectorArray,
        )
    ),
    model_info=PrefixedArray(
        Int32ul,
        Struct(
            str1=CString("utf-8"),
            elems=PrefixedArray(
                Int32ul,
                Struct(
                    float1=VectorArray,
                    float2=VectorArray,
                    float3=VectorArray,
                )
            )
        )
    ),
    strings_a=PrefixedArray(
        Int32ul,
        CString("utf-8"),
    ),
    unk_structs_a=PrefixedArray(
        Int32ul,
        Struct(
            str1=CString("utf-8"),
            char2=Byte,
            char3=Byte,
            char4=Byte,
            int5=Int32ul,
            int6=Int32ul,
            int7=Int32ul,
            char8=Byte,
            char9=Byte,
            int10=Int32ul,
            str11=PaddedString(16, "utf-8"),
            int12=Int32ul,
            float13=VectorArray,
            float14=VectorArray,
            float15=VectorArray,
            int16=Int32ul,
            float17=VectorArray,
        )
    ),
    strings_b=PrefixedArray(
        Int32ul,
        CString("utf-8"),
    ),
    unk_structs_b=PrefixedArray(
        Int32ul,
        Struct(
            str1=CString("utf-8"),
            int2=Int32ul,
            struct4=PrefixedArray(
                Int32ul,
                Struct(
                    int1=Int32ul,
                    long3=PrefixedArray(Int32ul, Int64ul),
                )
            )
        )
    )
)


class Bmssd(BaseResource):
    @classmethod
    def construct_class(cls, target_game: Game) -> Construct:
        return BMSSD



from mercury_engine_data_structures.formats.pkg import PKG
from mercury_engine_data_structures.game_check import Game
from test.test_lib import parse_and_build_compare


def test_compare_dread(dread_path):
    parse_and_build_compare(
        PKG, Game.DREAD, dread_path.joinpath("packs/system/system.pkg")
    )



import argparse
import asyncio
import itertools
import json
import logging
import typing
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Optional

from mercury_engine_data_structures import formats
from mercury_engine_data_structures.construct_extensions.json import convert_to_raw_python
from mercury_engine_data_structures.game_check import Game
from mercury_engine_data_structures.pkg_editor import PkgEditor


def game_argument_type(s: str) -> Game:
    try:
        return Game(int(s))
    except ValueError:
        # not a number, look by name
        for g in Game:
            g = typing.cast(Game, g)
            if g.name.lower() == s.lower():
                return g
        raise ValueError(f"No enum named {s} found")


def add_game_argument(parser: argparse.ArgumentParser, name="--game"):
    choices = []
    for g in Game:
        g = typing.cast(Game, g)
        choices.append(g.value)
        choices.append(g.name)

    parser.add_argument(name, help="The game of the file", type=game_argument_type, choices=list(Game), required=True)


def create_parser():
    parser = argparse.ArgumentParser()

    subparser = parser.add_subparsers(dest="command", required=True)

    decode = subparser.add_parser("decode")
    add_game_argument(decode)
    decode.add_argument("--format", help="Hint the format of the file. Defaults to extension.")
    decode.add_argument("--re-encode", help="Re-encode afterwards and compares to the original.", action="store_true")
    decode.add_argument("input_path", type=Path, help="Path to the file")

    replace_pkg_file = subparser.add_parser("replace-in-pkg")
    add_game_argument(replace_pkg_file)
    replace_pkg_file.add_argument("--pkg-input", type=Path, required=True, help="Path to the PKG file")
    replace_pkg_file.add_argument("--pkg-output", type=Path, required=True, help="Path to where write the updated PKG")
    replace_pkg_file.add_argument("--asset-id", type=int, required=True, help="Asset id to replace")
    replace_pkg_file.add_argument("asset_path", type=Path, help="Path to the updated asset")

    find_pkg = subparser.add_parser("find-pkg-for")
    find_pkg.add_argument("--root", type=Path, required=True, help="Path to the PKG files")
    group = find_pkg.add_mutually_exclusive_group(required=True)
    group.add_argument("--asset-name", type=str, help="Asset name to find")
    group.add_argument("--asset-id", type=int, help="Asset id to find")

    compare = subparser.add_parser("compare-files")
    add_game_argument(compare)
    compare.add_argument("--format", help="Hint the format of the file", required=True)
    compare.add_argument("--limit", help="Limit the number of files to test", type=int)
    compare.add_argument("input_path", type=Path, help="Path to the directory to glob")

    return parser


def dump_to(path: Path, decoded):
    def default(o):
        if callable(o):
            o = o()
        if isinstance(o, bytes):
            return len(o)

        raise TypeError(f"Object of type {o.__class__.__name__} is not JSON serializable")

    with path.open("w") as f:
        x = convert_to_raw_python(decoded)
        f.write(json.JSONEncoder(indent=4, default=default).encode(x))


def do_decode(args):
    input_path: Path = args.input_path
    file_format = args.format
    game: Game = args.game
    re_encode = args.re_encode

    if file_format is None:
        file_format = input_path.suffix[1:]

    resource_class = formats.format_for(file_format)

    raw = input_path.read_bytes()
    decoded_resource = resource_class.parse(raw, target_game=game)
    print(decoded_resource.raw)

    if re_encode:
        encoded = decoded_resource.build()
        if raw != encoded:
            print(f"{input_path}: Results differ (len(raw): {len(raw)}; len(encoded): {len(encoded)})")


def replace_in_pkg(args):
    game: Game = args.game
    input_pkg: Path = args.pkg_input
    output_pkg: Path = args.pkg_output
    asset_id: int = args.asset_id
    asset_path: Path = args.asset_path

    logging.info("Reading %s", input_pkg)
    input_bytes = input_pkg.read_bytes()

    logging.info("Parsing...")
    pkg = formats.Pkg.parse(input_bytes, target_game=game)

    logging.info("Reading %s", asset_path)
    new_file = asset_path.read_bytes()

    logging.info("Replacing asset in pkg")
    pkg.replace_asset(asset_id, new_file)

    logging.info("Building new pkg")
    encoded = pkg.build()

    logging.info("Writing new pkg to %s", output_pkg)
    output_pkg.parent.mkdir(parents=True, exist_ok=True)
    output_pkg.write_bytes(encoded)

    logging.info("Done")


def find_pkg_for(args):
    root: Path = args.root
    asset_id: int = args.asset_id
    asset_name: str = args.asset_name

    with PkgEditor.open_pkgs_at(root) as pkg_editor:
        pkg_editor = typing.cast(PkgEditor, pkg_editor)
        if asset_id is not None:
            items = list(pkg_editor.find_pkgs_for_asset_id(asset_id))
        else:
            items = list(pkg_editor.find_pkgs_for_name(asset_name))

        print(f"> Pkgs for {asset_id} / {asset_name}:")
        for it in items:
            print(it)


def decode_encode_compare_file(file_path: Path, game: Game, file_format: str):
    resource_class = formats.format_for(file_format)

    try:
        raw = file_path.read_bytes()
        resource = resource_class.parse(raw, target_game=game)
        encoded = resource.build()

        if raw != encoded and raw.rstrip(b"\xFF") != encoded:
            return f"{file_path}: Results differ (len(raw): {len(raw)}; len(encoded): {len(encoded)})"
        return None

    except Exception as e:
        return f"{file_path}: Received error - {e}"


async def compare_all_files_in_path(args):
    input_path: Path = args.input_path
    file_format: str = args.format
    game: Game = args.game
    limit: Optional[int] = args.limit

    def apply_limit(it):
        if limit is None:
            return it
        else:
            return itertools.islice(it, limit)

    loop = asyncio.get_running_loop()

    try:
        import tqdm
    except ImportError:
        tqdm = None

    errors = []

    with ProcessPoolExecutor(max_workers=1) as executor:
        files = apply_limit(input_path.rglob(f"*.{file_format.upper()}"))
        if tqdm is not None:
            files = tqdm.tqdm(files, unit=" file")

        results = [loop.run_in_executor(executor, decode_encode_compare_file, f, game, file_format) for f in files]
        as_completed = asyncio.as_completed(results)
        if tqdm is not None:
            as_completed = tqdm.tqdm(as_completed, total=len(results), unit=" file")

        for c in as_completed:
            message = await c
            if message:
                if tqdm is not None:
                    errors.append(message)
                    as_completed.set_postfix_str(f"{len(errors)} errors")
                else:
                    print(message)
            raise SystemExit

    if errors:
        print(f"{len(errors)} errors:")
        for m in errors:
            print(m)


def main():
    logging.basicConfig(level=logging.INFO)
    args = create_parser().parse_args()

    if args.command == "decode":
        do_decode(args)
    elif args.command == "replace-in-pkg":
        replace_in_pkg(args)
    elif args.command == "find-pkg-for":
        find_pkg_for(args)
    elif args.command == "compare-files":
        asyncio.run(compare_all_files_in_path(args))
    else:
        raise ValueError(f"Unknown command: {args.command}")



import typing
from typing import Optional

import construct
from construct import (
    Struct, PrefixedArray, Int64ul, Int32ul, Hex, Construct, Computed, Array, Tell,
    Aligned, FocusedSeq, Rebuild, Seek, Pointer, Prefixed, GreedyBytes,
)

from mercury_engine_data_structures import crc, hashed_names
from mercury_engine_data_structures.construct_extensions.alignment import AlignTo
from mercury_engine_data_structures.formats.base_resource import BaseResource, AssetId
from mercury_engine_data_structures.game_check import Game

Construct_AssetId = Hex(Int64ul)


def offset_for(con: Struct, name: str):
    result = 0
    for sc in con.subcons:
        sc = typing.cast(Construct, sc)
        if sc.name == name:
            return result
        result += sc.sizeof()
    raise construct.ConstructError(f"Unknown field: {name}")


def header_field(field_name: str):
    offset = offset_for(FileEntry, field_name)

    def result(ctx):
        parents = [ctx]
        while "_" in parents[-1]:
            parents.append(parents[-1]["_"])

        start_headers = None
        index = None

        for c in reversed(parents):
            if "_start_headers" in c:
                start_headers = c["_start_headers"]
                break

        for c in parents:
            if "_resource_index" in c:
                index = c["_resource_index"]
                break

        if index is None or start_headers is None:
            raise ValueError("Missing required context key")

        return start_headers + (index * FileEntry.sizeof()) + offset

    return result


FileEntry = Struct(
    asset_id=Construct_AssetId,
    start_offset=Int32ul,
    end_offset=Int32ul,
)

PKGHeader = Struct(
    header_size=Int32ul,
    data_section_size=Int32ul,
    file_entries=PrefixedArray(Int32ul, FileEntry),
)

PKG = Struct(
    _header_size=Int32ul,

    _data_section_size_address=Tell,
    _data_section_size=Int32ul,

    _num_files=Rebuild(Int32ul, construct.len_(construct.this.files)),
    _start_headers=Tell,
    _skip_headers=Seek(lambda ctx: ctx._num_files * FileEntry.sizeof(), 1),

    _align=AlignTo(128),
    _files_start=Tell,
    _update_header_size=Pointer(
        0x0,
        Rebuild(Int32ul, lambda ctx: ctx._files_start - Int32ul.sizeof()),
    ),
    files=Array(
        construct.this._num_files,
        Aligned(8, FocusedSeq(
            "item",
            _resource_index=Computed(lambda ctx: ctx["_index"]),

            actual_start_offset=Tell,
            start_offset=Pointer(header_field("start_offset"),
                                 Rebuild(Int32ul, lambda ctx: ctx.actual_start_offset)),
            end_offset=Pointer(header_field("end_offset"),
                               Rebuild(Int32ul, lambda ctx: ctx.start_offset + len(ctx.item.data))),
            item_size=Computed(lambda ctx: ctx.end_offset - ctx.start_offset),

            item=Struct(
                asset_id=Pointer(header_field("asset_id"), Construct_AssetId),
                asset_name=Computed(lambda ctx: hashed_names.name_for_asset_id(ctx.asset_id)),
                data=Prefixed(
                    Rebuild(
                        Computed(lambda ctx: ctx._.item_size),
                        construct.len_(construct.this.data),
                    ),
                    GreedyBytes,
                ),
            ),
        )),
    ),
    _files_end=Tell,
    _update_data_section_size=Pointer(
        lambda ctx: ctx._data_section_size_address,
        Rebuild(Int32ul, lambda ctx: ctx._files_end - ctx._files_start),
    ),
)


class Pkg(BaseResource):
    @classmethod
    def construct_class(cls, target_game: Game) -> Construct:
        return PKG

    def get_resource_by_asset_id(self, asset_id: AssetId) -> Optional[bytes]:
        for file in self.raw.files:
            if file.asset_id == asset_id:
                return file.data

    def get_resource_by_name(self, name: str) -> Optional[bytes]:
        return self.get_resource_by_asset_id(crc.crc64(name.encode("utf-8")))

    def replace_asset(self, asset_id: int, new_file: bytes):
        for file in self.raw.files:
            if file.asset_id == asset_id:
                file.data = new_file
                return
        raise ValueError(f"Unknown asset id: {asset_id}")



from typing import Any

from construct import EnumIntegerString
from construct.lib import ListContainer, Container


def convert_to_raw_python(value) -> Any:
    if callable(value):
        value = value()

    if isinstance(value, ListContainer):
        return [convert_to_raw_python(item) for item in value]

    if isinstance(value, Container):
        return {key: convert_to_raw_python(item) for key, item in value.items() if not key.startswith("_")}

    if isinstance(value, EnumIntegerString):
        return str(value)

    return value



from typing import Type

from mercury_engine_data_structures.formats.base_resource import BaseResource, AssetType
from mercury_engine_data_structures.formats.bmsad import Bmsad
from mercury_engine_data_structures.formats.bmscc import Bmscc
from mercury_engine_data_structures.formats.bmssd import Bmssd
from mercury_engine_data_structures.formats.brem import Brem
from mercury_engine_data_structures.formats.bres import Bres
from mercury_engine_data_structures.formats.brev import Brev
from mercury_engine_data_structures.formats.brfld import Brfld
from mercury_engine_data_structures.formats.brsa import Brsa
from mercury_engine_data_structures.formats.pkg import Pkg

ALL_FORMATS = {
    "PKG": Pkg,
    "BMSSD": Bmssd,
    "BMSAD": Bmsad,
    "BRFLD": Brfld,
    "BMSCC": Bmscc,
    "BRSA": Brsa,
    "BREM": Brem,
    "BRES": Bres,
    "BREV": Brev,
}


def format_for(type_name: AssetType) -> Type[BaseResource]:
    return ALL_FORMATS[type_name.upper()]



import copy
import typing

import construct
from construct import Adapter

StrId = construct.CString("utf-8")
Int: construct.FormatField = typing.cast(construct.FormatField, construct.Int32sl)
UInt: construct.FormatField = typing.cast(construct.FormatField, construct.Int32ul)
Float: construct.FormatField = typing.cast(construct.FormatField, construct.Float32l)
CVector2D = construct.Array(2, Float)
CVector3D = construct.Array(3, Float)
CVector4D = construct.Array(4, Float)


class DictAdapter(Adapter):
    def _decode(self, obj: construct.ListContainer, context, path):
        result = construct.Container()
        for item in obj:
            key = item[0]
            if key in result:
                raise construct.ConstructError(f"Key {key} found twice in object", path)
            result[key] = item[1]
        return result

    def _encode(self, obj: construct.Container, context, path):
        return construct.ListContainer(
            construct.ListContainer([type_, item])
            for type_, item in obj.items()
        )


def make_dict(value: construct.Construct):
    return DictAdapter(make_vector(construct.Sequence(StrId, value)))


def make_vector(value: construct.Construct):
    arr = construct.Array(
        construct.this.count,
        value,
    )
    arr.name = "items"

    def get_len(ctx):
        return len(ctx['items'])

    return construct.FocusedSeq(
        "items",
        "count" / construct.Rebuild(construct.Int32ul, get_len),
        arr,
    )


def make_enum(values: typing.Union[typing.List[str], typing.Dict[str, int]], *,
              add_invalid: bool = True):
    if isinstance(values, dict):
        mapping = copy.copy(values)
    else:
        mapping = {
            name: i
            for i, name in enumerate(values)
        }
    if add_invalid:
        mapping["Invalid"] = 0x7fffffff
    return construct.Enum(construct.Int32ul, **mapping)



import collections
import functools
import json
import math
import multiprocessing
import re
import traceback
import typing
from pathlib import Path

import ghidra_bridge

hash_str = "HashString"
register_field = "RegisterField"
prefixes_to_remove = [
    "(ObjectField *)",
    "&",
]

_aliases = {
    # weirdness
    "(undefined **)base::global::CFilePathStrId": "base::global::CFilePathStrId",
    "global::CStrId": "base::global::CStrId",
    "global::CFilePathStrId": "base::global::CFilePathStrId",
    "math::CVector3D": "base::math::CVector3D",

    # custom names
    "&DAT_7172642b18": "CGameLink<CActor>",
    "&DAT_717275c0d8": "CGameLink<CEntity>",
    "&DAT_7172642ed8": "base::global::CRntVector<CGameLink<CActor>>",
    "&DAT_717275c498": "base::global::CRntVector<CGameLink<CEntity>>",

    "&CGameLink_CActor_DAT_7172642b18": "CGameLink<CActor>",
    "&CGameLink<CEntity>::Serializer": "CGameLink<CEntity>",
    "&Vector_GameLink_CActor_7172642ed8": "base::global::CRntVector<CGameLink<CActor>>",
    "&Vector_CGameLink_CEntity_DAT_717275c498": "base::global::CRntVector<CGameLink<CEntity>>",

    "&Vector_PtrCTriggerLogicAction_DAT_71726f3930": "base::global::CRntVector<std::unique_ptr<CTriggerLogicAction>>",

    "&Vector_CXParasiteBehavior_71726c3030": "base::global::CRntVector<std::unique_ptr<CXParasiteBehavior>>",
    "&base::snd::ELowPassFilter_DAT_7108b13de8": "base::snd::ELowPassFilter",

    "&DAT_71726bb4c0": "base::global::CRntVector<CCentralUnitComponent::SStartPointInfo>",
    "&DAT_71726baee8": "base::global::CRntVector<std::unique_ptr<CCentralUnitWeightedEdges>>",
    "&DAT_71729a98a8": "base::global::CRntVector<SFallBackPath>",
    "&DAT_7172686f58": "base::global::CRntVector<std::unique_ptr<CEmmyOverrideDeathPositionDef>>",
    "&DAT_7172687378": "base::global::CRntVector<std::unique_ptr<CEmmyAutoForbiddenEdgesDef>>",
    "&DAT_7172687798": "base::global::CRntVector<std::unique_ptr<CEmmyAutoGlobalSmartLinkDef>>",
    "&DAT_71726ecbf0": "CFreezeRoomConfig",
    "&DAT_71726ecd30": "CFreezeRoomCoolConfig",
    "&DAT_71726ed380": "CHeatRoomConfig",
    "&DAT_71726ed4c0": "CHeatRoomCoolConfig",
    "&DAT_71726d53e0": "base::global::CRntVector<SBeamBoxActivatable>",
    "&vectSpawnPoints_DAT_71729aaf30": "base::global::CRntVector<CGameLink<CSpawnPointComponent>>",
    "&Vector_CSpawnerActorBlueprint_DAT_71729aa9d0": "base::global::CRntVector<CSpawnerActorBlueprint>",
    "&Trigger_DAT_71726f4968": "base::global::CRntVector<std::unique_ptr<CTriggerComponent::SActivationCondition>>",
    "&DictStr_ListStr_DAT_71726f5da0": "base::global::CRntDictionary<base::global::CStrId, base::global::CRntVector<base::global::CStrId>>",
    "&VectorStrId_DAT_7101d03998": "base::global::CRntVector<base::global::CStrId>",
    "&DAT_71726f8e78": "base::global::CRntVector<SDoorInfo>",
    "&DAT_71726fd0c0": "base::global::CRntVector<SWorldGraphNode>",
    "&DAT_71726d8090": "CDoorLifeComponent::SState",
    "&DAT_7101cf5c20": "base::core::CAssetLink",
    "&DAT_7101cf4aa8": "base::core::AssetID",
    "&SCameraSubRail_DAT_7172721790": "base::global::CRntVector<SCameraSubRail>",
    "&DAT_71726ee5e8": "base::global::CRntVector<EShinesparkTravellingDirection>",
    "&DAT_71726ee9e0": "base::global::CRntVector<ECoolShinesparkSituation>",
    "&Vector_STileInfo_71726b8960": "base::global::CRntVector<CBreakableTileGroupComponent::STileInfo>",
    "&DAT_7172721398": "CEditorRailSegment",
    "&DAT_71726efbb0": "base::global::CRntVector<DoorStateInfo>",
    "&DAT_7101d062b0": "base::global::CRntSmallDictionary<base::global::CStrId, base::global::CStrId>",
    "&DAT_7108b143d0": "base::spatial::CAABox",
    "&DAT_71729a2688": "base::global::CRntVector<SLogicSubPath>",
    "&DAT_71729a2290": "base::global::CRntVector<SLogicPathNode>",
}


def clean_crc_var(crc_var: str) -> str:
    for prefix in prefixes_to_remove:
        if crc_var.startswith(prefix):
            crc_var = crc_var[len(prefix):].strip()
    return crc_var


def get_field_registrations(bridge: ghidra_bridge.GhidraBridge, ifc, monitor, fields_function):
    res = bridge.remote_eval("""
        ifc.decompileFunction(fields_function, 180, monitor)
    """, timeout_override=200, fields_function=fields_function, ifc=ifc, monitor=monitor)

    decompiled_code = str(res.getCCodeMarkup())
    hash_call_re = re.compile(hash_str + r'\(([^,]+?),"?([^,]+?)"?,1\);')
    register_call_re = re.compile(register_field + r'\([^,]+?,([^,]+?),(.+?),([^,]+?),([^,]+?),([^,]+?)\);')

    crc_mapping = collections.defaultdict(list)
    fields = {}

    for m in hash_call_re.finditer(decompiled_code):
        crc_var, crc_string = m.group(1, 2)
        crc_mapping[clean_crc_var(crc_var)].append((m.start(), crc_string))

    for m in register_call_re.finditer(decompiled_code):
        crc_var, type_var = m.group(1, 2)

        offset = None
        crc_string = None
        for offset, crc_string in reversed(crc_mapping[clean_crc_var(crc_var)]):
            if offset < m.start():
                break

        if "&" in type_var:
            if "::_" in type_var:
                type_name = type_var[1:type_var.find("::_")]
            else:
                type_name = type_var
        else:
            i = decompiled_code.rfind(type_var, offset, m.start())
            end = decompiled_code.find(';', i)
            type_name = decompiled_code[i + len(type_var) + len(" = "):end]
            if type_name.endswith("::init()"):
                type_name = type_name[:-len("::init()")]

        fields[crc_string] = _aliases.get(type_name, type_name)

    return fields


def get_function_list() -> dict[str, tuple[int, int]]:
    with ghidra_bridge.GhidraBridge() as init_bridge:
        result = init_bridge.remote_eval("""
        [
            (f.getName(True), f.getID()) for f in currentProgram.getSymbolTable().getDefinedSymbols()
            if f.getName() == "fields" or f.getName() == "init"
        ]
        """)
        init_funcs = {}
        fields_funcs = {}
        for name, func_id in result:
            if not name.startswith("Reflection::"):
                continue
            name = name[len("Reflection::"):]

            if name.endswith("::init"):
                init_funcs[name[:-len("::init")]] = func_id
            elif name.endswith("::fields"):
                fields_funcs[name[:-len("::fields")]] = func_id

        return {
            name: (init_funcs.get(name), fields_funcs[name])
            for name in fields_funcs
        }


bridge: typing.Optional[ghidra_bridge.GhidraBridge] = None


def initialize_worker():
    global bridge, monitor, ifc
    bridge = ghidra_bridge.GhidraBridge(response_timeout=10)

    flat_api = bridge.get_flat_api()
    DecompileOptions = bridge.remote_import("ghidra.app.decompiler.DecompileOptions")
    DecompInterface = bridge.remote_import("ghidra.app.decompiler.DecompInterface")
    ConsoleTaskMonitor = bridge.remote_import("ghidra.util.task.ConsoleTaskMonitor")

    monitor = ConsoleTaskMonitor()
    ifc = DecompInterface()
    ifc.setOptions(DecompileOptions())
    ifc.openProgram(flat_api.currentProgram)


def decompile_type(type_name: str, init_id: typing.Optional[int], fields_id: int,
                   ) -> tuple[str, typing.Optional[str], dict[str, str]]:
    if bridge is None:
        raise ValueError("Bridge not initialized")

    bridge.remote_exec("""
def find_parent(f):
    super_namespace = f.getParentNamespace().getParentNamespace()
    for other in f.getCalledFunctions(None):
        if other.getName().startswith("init") and super_namespace != other.getParentNamespace():
            return other.getName(True)
    """)

    parent_init: typing.Optional[str] = None
    if init_id is not None:
        parent_init = bridge.remote_eval("""find_parent(
            currentProgram.getFunctionManager().getFunctionAt(
                currentProgram.getSymbolTable().getSymbol(func_id).getAddress()
            )
        )""", func_id=init_id)

    func = bridge.remote_eval("""
        currentProgram.getFunctionManager().getFunctionAt(
            currentProgram.getSymbolTable().getSymbol(func_id).getAddress()
        )
    """, func_id=fields_id)

    fields = get_field_registrations(bridge, ifc, monitor, func)

    if parent_init is not None:
        if parent_init.startswith("Reflection::"):
            parent_init = parent_init[len("Reflection::"):]
        parent_init = parent_init[:-len("::init")]

    return type_name, parent_init, fields


def decompile_in_background(all_fields_functions: dict[str, tuple[int, int]]):
    process_count = max(multiprocessing.cpu_count() - 2, 2)

    finished_count = 0
    failed = []

    total_count = len(all_fields_functions)
    num_digits = math.ceil(math.log10(total_count + 1))
    number_format = "[{0:" + str(num_digits) + "d}/{1}] "

    def report_update(msg: str):
        nonlocal finished_count
        finished_count += 1
        print(number_format.format(finished_count, total_count) + msg)

    result = {}

    def callback(r):
        type_name, parent, fields = r
        result[type_name] = {"parent": parent, "fields": fields}
        report_update(f"Parsed {type_name}")

    if total_count > process_count:
        with multiprocessing.Pool(processes=process_count, initializer=initialize_worker) as pool:
            def error_callback(name, e):
                failed.append(name)
                msg = "".join(traceback.format_exception(type(e), e, e.__traceback__))
                report_update(f"Failed {name}: {msg}")

            for n, f in all_fields_functions.items():
                pool.apply_async(
                    func=decompile_type,
                    args=(n, *f),
                    callback=callback,
                    error_callback=functools.partial(error_callback, n),
                )

            pool.close()
            pool.join()
    else:
        print("Less tasks than CPUs, just do it single-threaded.")
        failed.extend(all_fields_functions.keys())

    if failed:
        print(f"{len(failed)} function(s) failed, retrying on main thread.")
        initialize_worker()

    for n in failed:
        try:
            callback(decompile_type(n, *all_fields_functions[n]))
        except Exception as e:
            report_update(f"Failed {n}: {e}")

    return result


def main(only_missing: bool = True):
    print("Getting function list")
    all_fields_functions = get_function_list()
    print(f"Got {len(all_fields_functions)} functions!")

    path = Path(__file__).parent.joinpath("all_types.json")

    try:
        with path.open() as f:
            final_results = json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        final_results = {}

    if only_missing:
        for key in final_results.keys():
            all_fields_functions.pop(key, None)

    process_results = decompile_in_background(all_fields_functions)
    for key in sorted(process_results.keys()):
        final_results[key] = process_results[key]

    for data in final_results.values():
        for field in data["fields"].keys():
            value = data["fields"][field]
            if value in _aliases:
                value = _aliases[value]

            if not value.startswith("&"):
                if value.endswith("Ptr"):
                    value = value[:-len("Ptr")] + "*"
                value = value.replace("_const", " const")
                value = value.replace(",_", ", ")
                value = value.replace("Ptr>", "*>")

            data["fields"][field] = value

    with path.open("w") as f:
        json.dump({
            key: final_results[key]
            for key in sorted(final_results.keys())
        }, f, indent=4)


def simple_decompile():
    all_fields_functions = get_function_list()
    initialize_worker()

    func_name = "CComponent"
    print(decompile_type(func_name, *all_fields_functions[func_name]))

    # print(decompile_function(*all_fields_functions[4]))


if __name__ == '__main__':
    main()



from pathlib import Path

from construct.lib.containers import Container

from mercury_engine_data_structures.game_check import Game


def _parse_and_build_compare(module, game: Game, file_path: Path, print_data=False, save_file=None):
    raw = file_path.read_bytes()

    data = module.parse(raw, target_game=game)
    if print_data:
        print(data)
    encoded = module.build(data, target_game=game)

    if save_file:
        file_path.parent.joinpath(save_file).write_bytes(encoded)

    return raw, encoded, data


def parse_and_build_compare(module, game: Game, file_path: Path, print_data=False, save_file=None):
    raw, encoded, _ = _parse_and_build_compare(module, game, file_path, print_data, save_file)
    assert encoded == raw


def parse_and_build_compare_parsed(module, game: Game, file_path: Path, print_data=False, save_file=None):
    _, encoded, data = _parse_and_build_compare(module, game, file_path, print_data, save_file)

    data2 = module.parse(encoded, target_game=game)
    if print_data:
        print(data2)

    assert purge_hidden(data) == purge_hidden(data2)


def purge_hidden(data: Container) -> Container:
    data = {k: v for k, v in data.items() if not k.startswith("_")}
    return {k: purge_hidden(v) if isinstance(v, Container) else v for k, v in data.items()}



from typing import Dict, Union, Type

import construct
from construct import Construct, Int32ul, Probe, Struct, Adapter

from mercury_engine_data_structures import hashed_names
from mercury_engine_data_structures.common_types import make_vector
from mercury_engine_data_structures.construct_extensions.misc import ErrorWithMessage, ForceQuit
from mercury_engine_data_structures.hashed_names import PropertyEnum


def ConfirmType(name: str):
    def check(ctx):
        return ctx[f"{name}_type"] != name

    return construct.If(
        check,
        ErrorWithMessage(
            lambda ctx: f"Expected {name}, got {ctx[f'{name}_type']} ("
                        f"{hashed_names.all_property_id_to_name().get(ctx[f'{name}_type'])}) "
                        "without assigned type"
        ),
    )


class ObjectAdapter(Adapter):
    def _decode(self, obj: construct.ListContainer, context, path):
        result = construct.Container()
        for item in obj:
            if item.type in result:
                raise construct.ConstructError(f"Type {item.type} found twice in object", path)
            result[item.type] = item.item
        return result

    def _encode(self, obj: construct.Container, context, path):
        return construct.ListContainer(
            construct.Container(
                type=type_,
                item=item
            )
            for type_, item in obj.items()
        )


def Object(fields: Dict[str, Union[Construct, Type[Construct]]], *,
           debug=False) -> Construct:
    all_types = list(fields)

    fields = {
        name: name / conn
        for name, conn in fields.items()
    }
    for type_name in all_types:
        if type_name not in hashed_names.all_name_to_property_id():
            raise ValueError(f"Unknown type name: {type_name}, not in hashes database")

    switch = construct.Switch(
        construct.this.type,
        fields,
        ErrorWithMessage(
            lambda ctx: f"Type {ctx.type} not known, valid types are {all_types}."
        )
    )
    switch.name = "item"
    r = ObjectAdapter(make_vector(
        Struct(
            "type" / PropertyEnum,
            switch,
        )
    ))
    if debug:
        r.name = "fields"
        r = construct.FocusedSeq(
            "fields",
            r,
            "next_enum" / PropertyEnum,
            "probe" / Probe(lookahead=0x8),
            ForceQuit(),
        )

    return r



from construct import Adapter, Int32ub, Enum


class EnumAdapter(Adapter):
    def __init__(self, enum_class, subcon=Int32ub):
        super().__init__(Enum(subcon, enum_class))
        self._enum_class = enum_class

    def _decode(self, obj, context, path):
        return self._enum_class[obj]

    def _encode(self, obj, context, path):
        return obj.name



import construct
from construct import (
    Struct, Construct, Const, Int32ul, Hex, CString, Switch, Int16ul,
    PrefixedArray, Byte, Array, Float32l, Probe,
)

from mercury_engine_data_structures.construct_extensions.misc import ErrorWithMessage
from mercury_engine_data_structures.formats import BaseResource
from mercury_engine_data_structures.game_check import Game

Pickable = Struct(
    unknown=Int32ul,
)

component_types = {
    "PICKABLE": Pickable,
}

CCharClass = Struct(
    model_name=CString("utf-8"),
    unk_1=Int16ul,
    unk_2=Int32ul,
    unk_3=Int16ul,
    sub_actors=PrefixedArray(Int32ul, CString("utf-8")),
    unk_4=Array(9, Float32l),
    magic=Const(0xFFFFFFFF, Hex(Int32ul)),
    unk_5=Int16ul,
    unk_6=Byte,
    _=Probe(),
    components=PrefixedArray(
        Int32ul,
        Struct(
            type=CString("utf-8"),
            _=Probe(),
            component=Switch(
                construct.this.type,
                component_types,
                ErrorWithMessage("Unknown component type"),
            )
        ),
    ),
)

property_types = {
    "CCharClass": CCharClass,
}

BMSAD = Struct(
    magic_a=Const(b"MSAD"),
    magic_b=Const(0x0200000F, Hex(Int32ul)),
    name=CString("utf-8"),
    type=CString("utf-8"),
    property=Switch(
        construct.this.type,
        property_types,
        ErrorWithMessage("Unknown property type"),
    )
)


class Bmsad(BaseResource):
    @classmethod
    def construct_class(cls, target_game: Game) -> Construct:
        return BMSAD



import construct

from mercury_engine_data_structures.formats.dread_types import gameeditor_CGameModelRoot
from mercury_engine_data_structures.hashed_names import PropertyEnum


def create(name: str, magic_number: int):
    return construct.Struct(
        magic=construct.Const(name, PropertyEnum),
        magic_number=construct.Const(magic_number, construct.Hex(construct.Int32ul)),

        # gameeditor::CGameModelRoot
        root_type=construct.Const('Root', PropertyEnum),
        Root=gameeditor_CGameModelRoot,

        _end=construct.Terminated,
    )



import enum
from functools import partial

from construct import EnumIntegerString, IfThenElse, If


def get_version(this, enum_type):
    if "version" not in this:
        return get_version(this["_"], enum_type)
    else:
        if isinstance(this.version, EnumIntegerString):
            return int(this.version)
        if enum_type and isinstance(this.version, str):
            return enum_type[this.version]
        return this.version


def compare_version(version):
    if isinstance(version, enum.Enum):
        return partial(get_version, enum_type=type(version))
    return partial(get_version, enum_type=None)


def WithVersionElse(version, with_subcon, before_subcon):
    return IfThenElse(lambda this: compare_version(version)(this) >= version, with_subcon, before_subcon)


def WithVersion(version, subcon):
    return If(lambda this: compare_version(version)(this) >= version, subcon)


def BeforeVersion(version, subcon):
    return If(lambda this: compare_version(version)(this) < version, subcon)



"""
For checking which game is being parsed
"""
from enum import Enum
from typing import Any, Callable

from construct.core import IfThenElse


class Game(Enum):
    SAMUS_RETURNS = 10
    DREAD = 11

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


def get_current_game(ctx):
    result = ctx["_params"]["target_game"]
    if not isinstance(result, Game):
        raise ValueError(f"build/parse didn't set a valid target_game. Expected `Game`, got {result}")

    return result


def is_samus_returns(ctx):
    return get_current_game(ctx) == Game.SAMUS_RETURNS


def is_dread(ctx):
    return get_current_game(ctx) == Game.DREAD


def current_game_at_most(target: Game) -> Callable[[Any], bool]:
    def result(ctx):
        return get_current_game(ctx) <= target

    return result


def current_game_at_least(target: Game) -> Callable[[Any], bool]:
    def result(ctx):
        return get_current_game(ctx) >= target

    return result


def current_game_at_least_else(target: Game, subcon1, subcon2) -> IfThenElse:
    return IfThenElse(current_game_at_least(target), subcon1, subcon2)



import io

import construct
from construct import Construct, stream_tell, Subconstruct, GreedyBytes


class AlignTo(Construct):
    def __init__(self, modulus):
        super().__init__()
        self.modulus = modulus
        self.pattern = b"\x00"
        self.flagbuildnone = True

    def _parse(self, stream, context, path):
        modulus = construct.evaluate(self.modulus, context)
        position = stream_tell(stream, path)
        pad = modulus - (position % modulus)
        if pad < modulus:
            return construct.stream_read(stream, pad, path)
        return b""

    def _build(self, obj, stream, context, path):
        modulus = construct.evaluate(self.modulus, context)
        position = stream_tell(stream, path)
        pad = modulus - (position % modulus)
        if pad < modulus:
            construct.stream_write(stream, self.pattern * pad, pad, path)


class AlignedPrefixed(Subconstruct):
    def __init__(self, length_field, subcon, modulus, length_size, pad_byte=b"\xFF"):
        super().__init__(subcon)
        self.length_field = length_field
        self.modulus = modulus
        self.length_size = length_size
        self.pad_byte = pad_byte

    def _parse(self, stream, context, path):
        modulus = construct.evaluate(self.modulus, context)
        length_size = construct.evaluate(self.modulus, context)

        length = self.length_field._parsereport(stream, context, path)
        data = construct.stream_read(stream, length, path)
        pad = modulus - ((len(data) - length_size) % modulus)
        if pad < modulus:
            data += self.pad_byte * pad

        if self.subcon is construct.GreedyBytes:
            return data
        if type(self.subcon) is construct.GreedyString:
            return data.decode(self.subcon.encoding)
        return self.subcon._parsereport(io.BytesIO(data), context, path)

    def _build(self, obj, stream, context, path):
        modulus = construct.evaluate(self.modulus, context)
        length_size = construct.evaluate(self.modulus, context)

        stream2 = io.BytesIO()
        buildret = self.subcon._build(obj, stream2, context, path)
        data = stream2.getvalue()
        pad = modulus - ((len(data) - length_size) % modulus)
        if pad < modulus:
            data += self.pad_byte * pad

        self.length_field._build(len(data), stream, context, path)
        construct.stream_write(stream, data, len(data), path)
        return buildret

    def _sizeof(self, context, path):
        return self.length_field._sizeof(context, path) + self.subcon._sizeof(context, path)

    def _actualsize(self, stream, context, path):
        position1 = stream_tell(stream, path)
        length = self.length_field._parse(stream, context, path)
        position2 = stream_tell(stream, path)
        return (position2 - position1) + length


class PrefixedWithPaddingBefore(Subconstruct):
    def __init__(self, length_field, subcon, padding=32):
        super().__init__(subcon)
        self.padding = padding
        self.length_field = length_field

    def _parse(self, stream, context, path):
        length = self.length_field._parsereport(stream, context, path)
        bytes_to_pad = self.padding - (length % self.padding)
        if bytes_to_pad < self.padding:
            construct.stream_read(stream, bytes_to_pad, path)
        data = construct.stream_read(stream, length, path)
        if self.subcon is GreedyBytes:
            return data
        return self.subcon._parsereport(io.BytesIO(data), context, path)

    def _build(self, obj, stream, context, path):
        stream2 = io.BytesIO()
        buildret = self.subcon._build(obj, stream2, context, path)
        data = stream2.getvalue()
        length = len(data)
        self.length_field._build(length, stream, context, path)

        bytes_to_pad = self.padding - (length % self.padding)
        if bytes_to_pad < self.padding:
            construct.stream_write(stream, b"\x00" * bytes_to_pad, bytes_to_pad, path)

        construct.stream_write(stream, data, len(data), path)
        return buildret



from construct import Construct, Container

from mercury_engine_data_structures.game_check import Game


class BaseResource:
    _raw: Container
    target_game: Game

    def __init__(self, raw: Container, target_game: Game):
        self._raw = raw
        self.target_game = target_game

    @classmethod
    def construct_class(cls, target_game: Game) -> Construct:
        raise NotImplementedError()

    @classmethod
    def parse(cls, data: bytes, target_game: Game) -> "BaseResource":
        return cls(cls.construct_class(target_game).parse(data, target_game=target_game),
                   target_game)

    def build(self) -> bytes:
        return self.construct_class(self.target_game).build(self._raw, target_game=self.target_game)

    @property
    def raw(self) -> Container:
        return self._raw


AssetType = str
AssetId = int
